import os
from sys import getsizeof

from requests import HTTPError


def ensure_dir(path):
    if not os.path.isdir(path):
        os.makedirs(path)


def compute_memory_usage(obj):
    seen = set()
    stack = [obj]
    returned = 0
    while stack:
        obj = stack.pop()
        if id(obj) in seen:
            continue
        seen.add(id(obj))
        returned += getsizeof(obj)
        if isinstance(obj, dict):
            for key, value in obj.items():
                stack.append(key)
                stack.append(value)
        elif isinstance(obj, (list, tuple, set, frozenset)):
            stack.extend(obj)
    return returned


def raise_for_status(resp):
    try:
        resp.raise_for_status()
    except HTTPError as e:
        raise HTTPError(
            '{0.request.method} {0.request.url}: {0.status_code}\n\n{0.content}'.format(e.response),
            response=resp, request=resp.request)



from sentinels import NOTHING

from .contrib.utils import normalize_file_path
from .lazy_query import LazyQuery


class WarningContainer(object):

    def add_warning(self, message, filename=NOTHING, lineno=NOTHING, timestamp=NOTHING):
        return self.client.api.call_function('add_warning', {self._get_id_key(): self.id,
                                                             'message': message,
                                                             'lineno': lineno,
                                                             'filename': normalize_file_path(filename),
                                                             'timestamp': timestamp,
                                                            })

    def query_warnings(self):
        return LazyQuery(self.client, '/rest/warnings', query_params={self._get_id_key(): self.id})

    def _get_id_key(self):
        if type(self).__name__ == 'Test':
            return 'test_id'
        return 'session_id'



import copy
import gzip
import json
import random
import time

import requests
from requests.exceptions import ConnectionError, ReadTimeout

from munch import munchify
from sentinels import NOTHING
from urlobject import URLObject as URL

from .__version__ import __version__ as BACKSLASH_CLIENT_VERSION
from ._compat import BytesIO, TextIOWrapper, iteritems
from .comment import Comment
from .error import Error
from .exceptions import BackslashClientException, ParamsTooLarge
from .session import Session
from .suite import Suite
from .test import Test
from .user import User
from .utils import raise_for_status, compute_memory_usage
from .warning import Warning

_RETRY_STATUS_CODES = frozenset([
    requests.codes.bad_gateway,
    requests.codes.gateway_timeout,
])

_TYPES_BY_TYPENAME = {
    'session': Session,
    'test': Test,
    'error': Error,
    'warning': Warning,
    'comment': Comment,
    'suite': Suite,
    'user': User,
}


_COMPRESS_THRESHOLD = 4 * 1024
_MAX_PARAMS_COMPRESSED_SIZE = 5 * 1024 * 1024  # 5Mb
_MAX_PARAMS_UNCOMPRESSED_SIZE = 10 * 1024 * 1024 # 10Mb


class API(object):

    def __init__(self, client, url, runtoken):
        super(API, self).__init__()
        self.client = client
        self.url = URL(url)
        self.runtoken = runtoken
        self.session = requests.Session()
        self.session.headers.update({
            'X-Backslash-run-token': self.runtoken,
            'X-Backslash-client-version': BACKSLASH_CLIENT_VERSION,
        })
        self.call = CallProxy(self)
        self._cached_info = None

    def __del__(self):
        if self.session is not None:
            self.session.close()

    def info(self):
        """Inspects the remote API and returns information about its capabilities
        """
        if self._cached_info is None:
            resp = self.session.options(self.url.add_path('api'))
            raise_for_status(resp)
            self._cached_info = munchify(resp.json())
        return copy.deepcopy(self._cached_info)

    def call_function(self, name, params=None):
        is_compressed, data = self._serialize_params(params)
        headers = {'Content-type': 'application/json'}
        if is_compressed:
            headers['Content-encoding'] = 'gzip'

        for _ in self._iter_retries():

            try:
                resp = self.session.post(
                    self.url.add_path('api').add_path(name), data=data, headers=headers)
            except (ConnectionError, ReadTimeout, ):
                continue

            if resp.status_code not in _RETRY_STATUS_CODES:
                break
        else:
            raise BackslashClientException(
                'Maximum number of retries exceeded for calling Backslash API')

        raise_for_status(resp)

        return self._normalize_return_value(resp)

    def _iter_retries(self, timeout=30, sleep_range=(3, 10)):
        start_time = time.time()
        end_time = start_time + timeout
        while True:
            yield
            if time.time() < end_time:
                time.sleep(random.randrange(*sleep_range))

    def get(self, path, raw=False, params=None):
        resp = self.session.get(self.url.add_path(path), params=params)
        raise_for_status(resp)
        if raw:
            return resp.json()
        else:
            return self._normalize_return_value(resp)

    def delete(self, path, params=None):
        resp = self.session.delete(self.url.add_path(path), params=params)
        raise_for_status(resp)
        return resp

    def _normalize_return_value(self, response):
        json_res = response.json()
        if json_res is None:
            return None
        result = json_res.get('result')
        if result is None:
            if isinstance(json_res, dict):
                for key, value in json_res.items():
                    if isinstance(value, dict) and value.get('type') == key:
                        return self.build_api_object(value)
            return json_res
        elif isinstance(result, dict) and 'type' in result:
            return self.build_api_object(result)
        return result

    def build_api_object(self, result):
        objtype = self._get_objtype(result)
        if objtype is None:
            return result
        return objtype(self.client, result)

    def _get_objtype(self, json_object):
        typename = json_object['type']
        return _TYPES_BY_TYPENAME.get(typename)

    def _serialize_params(self, params):
        if params is None:
            params = {}

        returned = {}

        if compute_memory_usage(params) > _MAX_PARAMS_UNCOMPRESSED_SIZE:
            raise ParamsTooLarge()

        for param_name, param_value in iteritems(params):
            if param_value is NOTHING:
                continue
            returned[param_name] = param_value
        compressed = False
        returned = json.dumps(returned)
        if len(returned) > _COMPRESS_THRESHOLD:
            compressed = True
            returned = self._compress(returned)
        if len(returned) > _MAX_PARAMS_COMPRESSED_SIZE:
            raise ParamsTooLarge()
        return compressed, returned

    def _compress(self, data):
        s = BytesIO()

        with gzip.GzipFile(fileobj=s, mode='wb') as f:
            with TextIOWrapper(f) as w:
                w.write(data)

        return s.getvalue()


class CallProxy(object):

    def __init__(self, api):
        super(CallProxy, self).__init__()
        self._api = api

    def __getattr__(self, attr):
        if attr.startswith('_'):
            raise AttributeError(attr)

        def new_func(**params):
            return self._api.call_function(attr, params)

        new_func.__name__ = attr
        return new_func




class APIObject(object):

    def __init__(self, client, json_data):
        super(APIObject, self).__init__()
        self.client = client
        self._data = json_data

    @property
    def api_url(self):
        return self.client.url.add_path(self.api_path)

    @property
    def ui_url(self):
        raise NotImplementedError() # pragma: no cover

    def __eq__(self, other):
        if not isinstance(other, APIObject):
            return NotImplemented
        return self.client is other.client and self._data == other._data  # pylint: disable=protected-access

    def __ne__(self, other):
        return not (self == other)  # pylint: disable=superfluous-parens

    def __getattr__(self, name):
        try:
            return self.__dict__['_data'][name]
        except KeyError:
            raise AttributeError(name)

    def refresh(self):
        prev_id = self.id
        self._data = self._fetch()
        assert self.id == prev_id
        return self

    def _fetch(self):
        return self.client.api.get(self.api_path, raw=True)[self._data['type']]

    def __repr__(self):
        return '<API:{data[type]}:{data[id]}>'.format(data=self._data)

    def without_fields(self, field_names):
        new_data = dict((field_name, field_value)
                        for field_name, field_value in self._data.items()
                        if field_name not in field_names)
        return type(self)(self.client, new_data)



from __future__ import print_function

import functools
import hashlib
import itertools
import json
import os
import pkg_resources
import socket
import sys
import time
import webbrowser

import logbook
import requests
import vintage

try:
    import git
except Exception as e: # pylint: disable=broad-except
    pass

import slash
from sentinels import NOTHING
from slash import config as slash_config
from slash.plugins import PluginInterface, registers_on
from slash.utils.conf_utils import Cmdline, Doc
from urlobject import URLObject as URL
from requests import HTTPError
from .._compat import shellquote
from ..client import Backslash as BackslashClient
from ..exceptions import ParamsTooLarge
from ..utils import ensure_dir
from .keepalive_thread import KeepaliveThread
from .utils import normalize_file_path, distill_slash_traceback, distill_object_attributes, add_environment_variable_metadata
from ..lazy_query import LazyQuery
from ..session import APPEND_UPCOMING_TESTS_STR
from ..__version__ import __version__ as BACKSLASH_CLIENT_VERSION

_DEFAULT_CONFIG_FILENAME = os.path.expanduser('~/.backslash/config.json')

_logger = logbook.Logger(__name__)

_PWD = os.path.abspath('.')

_HAS_TEST_AVOIDED = (int(slash.__version__.split('.')[0]) >= 1)
_HAS_SESSION_INTERRUPT = hasattr(slash.hooks, 'session_interrupt')
_HAS_TEST_DISTRIBUTED = hasattr(slash.hooks, 'test_distributed')
_HAS_APP_QUIT = hasattr(slash.hooks, 'app_quit')

def handle_exceptions(func):

    @functools.wraps(func)
    def new_func(self, *args, **kwargs):
        try:
            return func(self, *args, **kwargs)
        except Exception: # pylint: disable=broad-except
            exc_info = sys.exc_info()
            if not self._handle_exception(exc_info) or self._propagate_exceptions: # pylint: disable=protected-access
                raise
    return new_func


class BackslashPlugin(PluginInterface):

    client = current_test = session = None

    def __init__(self, url=None, keepalive_interval=None, runtoken=None,
                 propagate_exceptions=False, config_filename=_DEFAULT_CONFIG_FILENAME):
        super(BackslashPlugin, self).__init__()
        self._url = url
        self._repo_cache = {}
        self._config_filename = config_filename
        self._file_hash_cache = {}
        self._keepalive_interval = keepalive_interval
        self._keepalive_thread = None
        self._error_containers = {}
        self._runtoken = runtoken
        self._propagate_exceptions = propagate_exceptions
        self._started = False

    @property
    def rest_url(self):
        if self.client is None:
            return None
        return self.client.url.add_path('rest')

    @property
    def webapp_url(self):
        if self.client is None:
            return None
        return self.client.get_ui_url()

    @property
    def session_webapp_url(self):
        session = slash.context.session
        if session is None or self.client is None:
            return None
        return self.client.get_ui_url('sessions/{}'.format(session.id))

    def _handle_exception(self, exc_info):
        pass

    def _handle_keepalive_exception(self, exc_info):
        pass

    def _get_backslash_url(self):
        return self._url

    def get_name(self):
        return 'backslash'

    def get_default_config(self):
        return {
            "session_ttl_days": 0 // Doc(
                'Optional number of days after which this session will be discarded '
                'from Backslash') // Cmdline(arg='--session-ttl-days', metavar='DAYS'),


            "session_labels": [] // Doc('Specify labels to be added to the session when reported') \
                                 // Cmdline(append="--session-label", metavar="LABEL"),
            "blacklisted_warnings_category": [] // Doc(
                'Specify warnings categories which should not be reported to backslash'),
            "report_test_docstrings": False // Doc(
                'Add test docstring to backslash test metadata') // Cmdline(on="--report_test_docstrings"),
        }

    @handle_exceptions
    def activate(self):
        if self._runtoken is None:
            self._runtoken = self._ensure_run_token()
        self.client = BackslashClient(URL(self._get_backslash_url()), self._runtoken)

    def deactivate(self):
        if self._keepalive_thread is not None:
            self._keepalive_thread.stop()
        super(BackslashPlugin, self).deactivate()

    def _notify_session_start(self):
        metadata = self._get_initial_session_metadata()
        is_parent_session = False
        parent_logical_id = child_id = None
        is_slash_support_parallel = getattr(slash_config.root, 'parallel', False)
        if is_slash_support_parallel and slash_config.root.parallel.num_workers:
            child_id = slash_config.root.parallel.worker_id
            if child_id is not None:
                parent_logical_id = getattr(slash.context.session, 'parent_session_id', None)
            else:
                is_parent_session = True

        self.session = self.client.report_session_start(
            logical_id=slash.context.session.id,
            parent_logical_id=parent_logical_id,
            is_parent_session=is_parent_session,
            child_id=child_id,
            total_num_tests=slash.context.session.get_total_num_tests(),
            hostname=socket.getfqdn(),
            keepalive_interval=self._keepalive_interval,
            infrastructure='slash',
            metadata=metadata,
            **self._get_extra_session_start_kwargs()
        )
        self._started = True
        for warning in slash.context.session.warnings:
            self.warning_added(warning)
        for label in self.current_config.session_labels:
            self.client.api.call.add_label(session_id=self.session.id, label=label)

        if self._keepalive_interval is not None:
            self._keepalive_thread = KeepaliveThread(
                self.client,
                self.session,
                self._keepalive_interval,
                error_callback=self._handle_keepalive_exception
            )
            self._keepalive_thread.start()

    @handle_exceptions
    def session_start(self):
        if self.session is None:
            self._notify_session_start()

    def _get_initial_session_metadata(self):
        returned = {
            'slash::version': slash.__version__,
            'slash::commandline': ' '.join(shellquote(arg) for arg in sys.argv),
            'backslash_client_version': BACKSLASH_CLIENT_VERSION,
            'python_version': '.'.join(map(str, sys.version_info[:3])),
            'process_id': os.getpid(),
        }

        add_environment_variable_metadata(metadata=returned)
        return returned

    def _get_extra_session_start_kwargs(self):
        returned = {}
        ttl_seconds = self.current_config.session_ttl_days * 24 * 60 * 60
        if ttl_seconds:
            returned['ttl_seconds'] = ttl_seconds
        return returned

    @slash.plugins.register_if(_HAS_TEST_AVOIDED)
    @handle_exceptions
    def test_avoided(self, reason):
        self.test_start()
        self.test_skip(reason=reason)
        self.test_end()


    @handle_exceptions
    def test_interrupt(self):
        if self.current_test is not None:
            self.current_test.report_interrupted()

    @slash.plugins.register_if(_HAS_SESSION_INTERRUPT)
    @handle_exceptions
    def session_interrupt(self):
        if self.session is not None:
            self.session.report_interrupted()

    @slash.plugins.registers_on(None)
    @handle_exceptions
    def report_planned_tests(self, tests):
        if not APPEND_UPCOMING_TESTS_STR in self.client.api.info().endpoints:
            return

        tests_metadata = []
        for test in tests:
            test_info = self._get_test_info(test)
            current = {'test_logical_id':test.__slash__.id,
                       'file_name':test_info['file_name'],
                       'name':test_info['name'],
                       'class_name':test_info['class_name']
                      }
            if 'variation' in test_info:
                current['variation'] = test_info['variation']
            tests_metadata.append(current)

        tests_count = 0
        batch_size = 100
        try:
            while tests_count < len(tests_metadata):
                self.session.report_upcoming_tests(tests_metadata[tests_count:tests_count+batch_size])
                tests_count += batch_size
        except requests.exceptions.HTTPError:
            _logger.error('Ignoring exception while reporting planned tests', exc_info=True)

    @handle_exceptions
    def test_start(self):
        kwargs = self._get_test_info(slash.context.test)
        self._update_scm_info(kwargs)
        tags = slash.context.test.__slash__.tags
        tag_dict = {tag_name: tags[tag_name] for tag_name in tags}

        if tag_dict:
            kwargs['metadata'] = {
                'slash::tags': {
                    'values': {tag_name: tag_value for tag_name, tag_value in tag_dict.items() if tag_value is not NOTHING},
                    'names': list(tag_dict),
                },
            }

        log_path = slash.context.result.get_log_path()
        if log_path:
            kwargs.setdefault('metadata', {})['local_log_path'] = os.path.abspath(log_path)

        if self.current_config.report_test_docstrings and slash.test.get_test_function().__doc__:
            kwargs.setdefault('metadata', {})['docstring'] = slash.test.get_test_function().__doc__

        self.current_test = self.session.report_test_start(
            test_logical_id=slash.context.test.__slash__.id,
            test_index=slash.context.test.__slash__.test_index1,
            **kwargs
        )
        self._error_containers[slash.context.test.__slash__.id] = self.current_test

    @slash.plugins.register_if(_HAS_TEST_DISTRIBUTED)
    @handle_exceptions #pylint: disable=unused-argument
    def test_distributed(self, test_logical_id, worker_session_id): #pylint: disable=unused-argument
        if 'report_test_distributed' in self.client.api.info().endpoints:
            self.current_test = self.session.report_test_distributed(test_logical_id)


    @handle_exceptions
    def test_skip(self, reason=None):
        self.current_test.mark_skipped(reason=reason)

    @slash.plugins.registers_on(None)
    def is_session_exist(self, session_id):
        try:
            self.client.api.get('/rest/sessions/{0}'.format(session_id))
            return True
        except HTTPError as e:
            if e.response.status_code == 404:
                return False
            raise

    @handle_exceptions
    @slash.plugins.registers_on(None)
    def get_tests_to_resume(self, session_id, filters_dict):
        """Queries backslash specific session's tests
        :param session_id: the wanted session
        :param filters_dict: a dictionary containing filters for backslash tests query
        :rtype: list of test objects
        """
        max_retries = 3
        for i in range(max_retries):
            try:
                default_params_dict = {x: 'true' for x in ['show_planned', 'show_skipped', 'show_unsuccessful', 'show_abandoned']}
                default_params_dict.update({'session_id': session_id, 'show_successful': 'false'})
                default_params_dict.update(filters_dict)
                return reversed(LazyQuery(self.client, '/rest/tests', query_params=default_params_dict).all())
            except HTTPError:
                if i == max_retries-1:
                    raise

    def _get_test_info(self, test):
        if test.__slash__.is_interactive() and \
           pkg_resources.parse_version(slash.__version__) < pkg_resources.parse_version('1.6.0'):
            returned = {
                'file_name': '<interactive>',
                'class_name': '<interactive>',
                'name': '<interactive>',
                'is_interactive': True
            }
        else:
            test_display_name = test.__slash__.address
            if set(test_display_name) & set('/.'):
                test_display_name = test.__slash__.function_name

            returned = {
                'file_name': normalize_file_path(test.__slash__.file_path),
                'class_name': test.__slash__.class_name,
                'name': test_display_name,
                'is_interactive': test.__slash__.is_interactive(),
                }
        variation = getattr(test.__slash__, 'variation', None)
        if variation:
            if hasattr(test.__slash__.variation, 'labels'):
                items = test.__slash__.variation.labels.items()
                returned['parameters'] = variation.values.copy()
            elif hasattr(test.__slash__.variation, 'id'):
                items = test.__slash__.variation.id.items()
                returned['parameters'] = variation.values.copy()
            else:
                items = test.__slash__.variation.items()
            returned['variation'] = dict((name, value) for name, value in items)

        return returned

    def _update_scm_info(self, test_info):
        try:
            test_info['file_hash'] = self._calculate_file_hash(test_info['file_name'])
            dirname = os.path.dirname(test_info['file_name'])
            repo = self._repo_cache.get(dirname, NOTHING)
            if repo is NOTHING:
                repo = self._repo_cache[dirname] = self._get_git_repo(dirname)
            if repo is None:
                return
            test_info['scm'] = 'git'
            try:
                hexsha = repo.head.commit.hexsha
            except Exception: # pylint: disable=broad-except
                _logger.debug('Unable to get commit hash', exc_info=True)
                hexsha = None
            test_info['scm_revision'] = hexsha
            test_info['scm_dirty'] = bool(repo.untracked_files or repo.index.diff(None) or repo.index.diff(repo.head.commit))
            if self.client.api.info().endpoints.report_test_start.version >= 3:
                if not repo.head.is_detached:
                    test_info['scm_local_branch'] = repo.active_branch.name
                    tracking_branch = repo.active_branch.tracking_branch()
                    if tracking_branch is not None:
                        test_info['scm_remote_branch'] = tracking_branch.name
        except Exception: # pylint: disable=broad-except
            _logger.warning('Error when obtaining SCM information', exc_info=True)

    def _calculate_file_hash(self, filename):
        returned = self._file_hash_cache.get(filename)
        if returned is None:
            try:
                with open(filename, 'rb') as f:
                    data = f.read()
                    h = hashlib.sha1()
                    h.update('blob '.encode('utf-8'))
                    h.update('{0}\0'.format(len(data)).encode('utf-8'))
                    h.update(data)
            except IOError as e:
                _logger.debug('Ignoring IOError {0!r} when calculating file hash for {1}', e, filename)
                returned = None
            else:
                returned = h.hexdigest()
            self._file_hash_cache[filename] = returned

        return returned

    def _get_git_repo(self, dirname):
        if not os.path.isabs(dirname):
            dirname = os.path.abspath(os.path.join(_PWD, dirname))
        while dirname != '/':
            if os.path.isdir(os.path.join(dirname, '.git')):
                return git.Repo(dirname)
            dirname = os.path.normpath(os.path.abspath(os.path.join(dirname, '..')))
        return None

    @handle_exceptions
    def test_end(self):
        if self.current_test is None:
            return

        details = {}

        if hasattr(slash.context.result, 'details'):
            additional = slash.context.result.details.all()
        else:
            additional = slash.context.result.get_additional_details()
        details.update(additional)
        self.current_test.set_metadata_dict(details)
        self.current_test.report_end()
        self.current_test = None

    @handle_exceptions
    def session_end(self):
        self._session_report_end('session_end')

    @slash.plugins.register_if(_HAS_APP_QUIT)
    @handle_exceptions  # pylint: disable=unused-argument
    def app_quit(self):
        self._session_report_end('app_quit')

    def _session_report_end(self, hook_name):
        if not self._started:
            return
        try:
            if self._keepalive_thread is not None:
                self._keepalive_thread.stop()

            kwargs = {}
            session_results = getattr(slash.session, 'results', None)
            has_fatal_errors = hasattr(session_results, 'has_fatal_errors') and session_results.has_fatal_errors()
            if self.client.api.info().endpoints.report_session_end.version >= 2:
                kwargs['has_fatal_errors'] = has_fatal_errors
            self.session.report_end(**kwargs)
            self._started = False
        except Exception:       # pylint: disable=broad-except
            _logger.error('Exception ignored in {}'.format(hook_name), exc_info=True)

    @handle_exceptions
    def error_added(self, result, error):
        self._add_exception(result=result, exception=error)

    @slash.plugins.register_if(hasattr(slash.hooks, 'interruption_added'))
    @handle_exceptions
    def interruption_added(self, result, exception):
        self._add_exception(result=result, exception=exception, is_interruption=True)


    def _add_exception(self, result, exception, is_interruption=False):
        has_interruptions = self.client.api.info().endpoints.add_error.version >= 4
        if is_interruption and not has_interruptions:
            _logger.debug('Server does not support recording is_interruption exceptions. Skipping reporting')
            return

        if result is slash.session.results.global_result:
            error_container = self.session
        else:
            error_container = self._error_containers.get(result.test_metadata.id, self.current_test) or self.session

        if error_container is None:
            _logger.debug('Could not determine error container to report on for {}', result)
            return

        with vintage.get_no_deprecations_context():
            exception_attrs = getattr(exception, 'exception_attributes', NOTHING)

        if exception_attrs is NOTHING and hasattr(exception, 'exc_info'):
            exception_attrs = distill_object_attributes(exception.exc_info[1])
        kwargs = {'exception_type': exception.exception_type.__name__ if exception.exception_type is not None else None,
                  'traceback': distill_slash_traceback(exception), 'exception_attrs': exception_attrs}
        if exception.message:
            message = exception.message
        elif hasattr(exception, 'exception_str'):
            message = exception.exception_str
        else:
            message = str(exception.exception)


        kwargs['message'] = message

        if has_interruptions:
            kwargs['is_interruption'] = is_interruption

        for compact_variables in [False, True]:
            if compact_variables:
                for frame in kwargs['traceback']:
                    frame['globals'] = None
                    frame['locals'] = None
            try:
                error_container.add_error(**kwargs)
            except ParamsTooLarge:
                if compact_variables:
                    raise
                # continue to try compacting
            else:
                break

    @handle_exceptions
    def warning_added(self, warning):
        if any(issubclass(warning.category, b_cls) for b_cls in self.current_config.blacklisted_warnings_category):
            return
        kwargs = {'message': warning.message, 'filename': warning.filename, 'lineno': warning.lineno}
        warning_obj = self.current_test if self.current_test is not None else self.session
        if warning_obj is not None:
            warning_obj.add_warning(**kwargs)

    @handle_exceptions
    def exception_caught_before_debugger(self, **_):
        if self.session is not None and slash.config.root.debug.enabled:
            self.session.report_in_pdb()

    @handle_exceptions
    def exception_caught_after_debugger(self, **_):
        if self.session is not None and slash.config.root.debug.enabled:
            self.session.report_not_in_pdb()

    #### Token Setup #########
    def _ensure_run_token(self):

        if self._runtoken is None:

            tokens = self._get_existing_tokens()

            self._runtoken = tokens.get(self._get_backslash_url())
            if self._runtoken is None:
                self._runtoken = self._fetch_token()
                self._save_token(self._runtoken)

        return self._runtoken

    def _get_existing_tokens(self):
        return self._get_config().get('run_tokens', {})

    def _get_config(self):
        if not os.path.isfile(self._config_filename):
            return {}
        with open(self._config_filename) as f:
            return json.load(f)

    def _save_token(self, token):
        tmp_filename = self._config_filename + '.tmp'
        cfg = self._get_config()
        cfg.setdefault('run_tokens', {})[self._get_backslash_url()] = token

        ensure_dir(os.path.dirname(tmp_filename))

        with open(tmp_filename, 'w') as f:
            json.dump(cfg, f, indent=2)
        os.rename(tmp_filename, self._config_filename)

    @registers_on(None)
    def fetch_token(self, username, password):
        url = URL(self._get_backslash_url())
        with requests.Session() as s:
            resp = s.get(self._get_token_request_url())
            resp.raise_for_status()
            response_url = resp.json()['url']
            request_id = response_url.split('/')[-1]
            s.post(url.add_path('login'),
                   data=json.dumps({'username': username, 'password': password}),
                   headers={'Content-type': 'application/json'})\
             .raise_for_status()

            s.post(URL(self._get_backslash_url()).add_path('/runtoken/request/{}/complete'.format(request_id)))\
             .raise_for_status()

            resp = s.get(response_url)
            resp.raise_for_status()
            returned = self._runtoken = resp.json()['token']
            return returned


    def _fetch_token(self):
        """Template method for the fallback behavior of fetching a runtoken from the server.
        By default this uses _fetch_token_via_browser to initiate the browser-based
        authentication.

        You can override this method to provide a custom logic for fetching your tokens
        """
        return self._fetch_token_via_browser()


    def _fetch_token_via_browser(self):
        opened_browser = False
        url = self._get_token_request_url()
        for retry in itertools.count():
            resp = requests.get(url)
            resp.raise_for_status()
            data = resp.json()
            if retry == 0:
                url = data['url']
            token = data.get('token')
            if token:
                return token
            if not opened_browser:
                if not self._browse_url(data['complete']):
                    print('Could not open browser to fetch user token. Please login at', data['complete'])
                print('Waiting for Backlash token...')
                opened_browser = True
            time.sleep(1)

    def _get_token_request_url(self):
        return URL(self._get_backslash_url()).add_path('/runtoken/request/new')

    def _browse_url(self, url):
        if 'linux' in sys.platform and os.environ.get('DISPLAY') is None:
            return False # can't start browser
        return webbrowser.open_new(url)



import logbook
import sys
import threading

_logger = logbook.Logger(__name__)


class KeepaliveThread(threading.Thread):

    def __init__(self, client, session, interval, error_callback=None):
        super(KeepaliveThread, self).__init__()
        self._client = client
        self._session = session
        self._interval = interval / 2.0
        self._stopped_event = threading.Event()
        self._error_callback = error_callback
        self.daemon = True

    def run(self):
        _logger.debug('Backslash keepalive thread started')
        try:
            while not self._stopped_event.is_set():
                self._stopped_event.wait(timeout=self._interval)
                self._session.send_keepalive()
        except Exception: #pylint: disable=broad-except
            _logger.error('Quitting keepalive thread due to exception', exc_info=True)
            if self._error_callback is not None:
                self._error_callback(sys.exc_info())
            raise
        finally:
            _logger.debug('Backslash keepalive thread terminated')

    def stop(self):
        self._stopped_event.set()



from backslash.utils import compute_memory_usage

def test_getsizeof():
    large_object = {'x': (1, 2, 3), 'y': [1, 2, 3, 4, 5], 'd': {'key': 'hey', 'key2': {}, 'key3': ["", {}, ()]}}
    assert compute_memory_usage(large_object) > large_object.__sizeof__()



class TimingContainer(object):

    def report_timing_start(self, name):
        self._report('start', name)

    def report_timing_end(self, name):
        self._report('end', name)

    def _report(self, start_stop, name):
        kwargs = {'name': name}
        kwargs.update(self._get_identity_kwargs())
        self.client.api.call_function('report_timing_{}'.format(start_stop), kwargs) # pylint: disable=no-member

    def _get_identity_kwargs(self):
        if self.type.lower() == 'session': # pylint: disable=no-member
            return {'session_id': self.id} # pylint: disable=no-member

        return {'test_id': self.id, 'session_id': self.session_id} # pylint: disable=no-member

    def get_timings(self):
        return self.client.api.call.get_timings(**self._get_identity_kwargs()) # pylint: disable=no-member



import collections
import itertools

from sentinels import NOTHING

from ._compat import iteritems
from .utils import raise_for_status


class LazyQuery(object):

    def __init__(self, client, path=None, url=None, query_params=None, page_size=100):
        super(LazyQuery, self).__init__()
        self._client = client
        if url is None:
            url = client.api.url
        if path is not None:
            url = url.add_path(path)
        if query_params is not None:
            for (param, value) in query_params.items():
                url = url.add_query_param(param, str(value))
        self._url = url
        self._fetched = collections.defaultdict(lambda: NOTHING)
        self._total_num_objects = None
        self._page_size = page_size
        self._typename = None

    def all(self):
        return list(self)

    def filter(self, *filter_objects, **fields):
        returned_url = self._url
        for filter_object in filter_objects:
            returned_url = filter_object.add_to_url(returned_url)
        for field_name, field_value in iteritems(fields):
            returned_url = returned_url.add_query_param(field_name, str(field_value))
        return LazyQuery(self._client, url=returned_url, page_size=self._page_size)

    def __repr__(self):
        return '<Query {0!r}>'.format(str(self._url))

    def __iter__(self):
        for i in itertools.count():
            item = self._fetch_index(i)
            if item is NOTHING:
                break
            yield item

    def __getitem__(self, idx):
        if isinstance(idx, slice):
            raise NotImplementedError() # pragma: no cover
        if idx < 0:
            if idx < -len(self):
                raise IndexError()
            idx = len(self) + idx
        returned = self._fetch_index(idx)
        if returned is NOTHING:
            raise IndexError(idx)
        return returned

    def _fetch_index(self, index):
        returned = self._fetched[index]
        if returned is NOTHING:
            page_index = (index // self._page_size) + 1
            self._fetch_page(page_index)
            returned = self._fetched[index]
        return returned

    def _fetch_page(self, page_index):
        assert page_index != 0
        response = self._client.api.session.get(self._url.add_query_param(
            'page', str(page_index)).add_query_param('page_size', str(self._page_size)))
        raise_for_status(response)
        response_data = response.json()
        keys = [key for key in response_data if key != 'meta']
        if len(keys) > 1:
            raise RuntimeError('Multiple keys returned')
        [obj_typename] = keys
        if self._typename is not None and obj_typename != self._typename:
            raise RuntimeError('Got different typename in query: {!r} != {!r}'.format(obj_typename, self._typename))
        self._typename = obj_typename

        for index, json_obj in enumerate(response_data[self._typename]):
            real_index = ((page_index - 1) * self._page_size) + index
            self._fetched[real_index] = self._client.api.build_api_object(json_obj)
        return response_data

    def count(self):
        self._fetch_index(0)
        return self._total_num_objects



from uuid import uuid4
import pytest
from backslash import Backslash
from backslash.api_object import APIObject
from backslash import test, session

# pylint: disable=redefined-outer-name

def test_api_object_equality(client, data1):
    assert APIObject(client, data1) == APIObject(client, data1)
    assert not APIObject(client, data1) != APIObject(client, data1)  # pylint: disable=unneeded-not

def test_api_object_equality_reflexive(client, data1):
    api_object = APIObject(client, data1)
    assert api_object == api_object
    assert not api_object != api_object  # pylint: disable=unneeded-not

def test_api_object_not_equal_when_different_clients(data1):
    a = APIObject(object(), data1)
    b = APIObject(object(), data1)
    assert not a == b  # pylint: disable=unneeded-not
    assert a != b

def test_api_object_not_equal_different_data(client, data1, data2):
    a = APIObject(client, data1)
    b = APIObject(client, data2)
    assert not a == b  # pylint: disable=unneeded-not
    assert a != b

def test_object_api_url(client):
    data = {'api_path': '/rest/objects/1'}
    obj = APIObject(client, data)
    assert obj.api_url == 'http://127.0.0.1:12345/rest/objects/1'

def test_object_ui_url(client):
    obj = APIObject(client, {})
    with pytest.raises(NotImplementedError):
        obj.ui_url  # pylint: disable=pointless-statement

@pytest.mark.parametrize('object_type', [test.Test, session.Session])
@pytest.mark.parametrize('use_logical', [True, False])
def test_ui_url(client, object_type, logical_id, use_logical):
    id = 1002
    data = {'id': id, 'logical_id': logical_id if use_logical else None}
    if object_type is test.Test:
        data['session_display_id'] = str(uuid4())
    obj = object_type(client, data)

    url = obj.ui_url
    display_id = logical_id if use_logical else id
    if object_type is test.Test:
        assert url == client.url + '/#/sessions/{}/tests/{}'.format(data['session_display_id'], display_id)
    else:
        assert url == client.url + '/#/{}s/{}'.format(object_type.__name__.lower(), display_id)


@pytest.fixture
def logical_id():
    return str(uuid4())

@pytest.fixture
def client():
    return Backslash('http://127.0.0.1:12345', runtoken=None)



@pytest.fixture
def data1():
    return {'name': 'data1'}

@pytest.fixture
def data2():
    return {'name': 'data2'}



import gzip
import json
import tempfile

from ._compat import TextIOWrapper

import logbook
from sentinels import NOTHING


_logger = logbook.Logger(__name__)


class ErrorContainer(object):

    def add_error(self, message, exception_type=NOTHING, traceback=NOTHING,
                  timestamp=NOTHING, is_failure=NOTHING, exception_attrs=NOTHING, is_interruption=NOTHING):

        kwargs = {self._get_id_key(): self.id,  # pylint: disable=no-member
                  'message': message,
                  'exception_type': exception_type,
                  'is_failure': is_failure,
                  'timestamp': timestamp,
                  'is_interruption': is_interruption,
                  }

        add_error_version = self.client.api.info().endpoints.add_error.version # pylint: disable=no-member
        has_streaming_upload = add_error_version >= 2
        has_exception_attrs = add_error_version >= 3

        if not has_streaming_upload:
            if traceback is not NOTHING:
                kwargs['traceback'] = traceback
            traceback_info = NOTHING
        else:
            if has_exception_attrs:
                traceback_info = {'traceback': None if traceback is NOTHING else traceback,
                                  'exception': {
                                      'attributes': None if  exception_attrs is NOTHING else exception_attrs
                                  }
                }
            else:
                traceback_info = traceback

        returned = self.client.api.call_function('add_error', kwargs) # pylint: disable=no-member

        if has_streaming_upload and traceback_info is not NOTHING:
            self._compress_traceback(returned, traceback_info)

        return returned

    def _compress_traceback(self, error, traceback_info):
        traceback_url = error.api_url.add_path('traceback')
        with tempfile.TemporaryFile(mode='w+b') as traceback_file:
            try:
                with gzip.GzipFile(fileobj=traceback_file, mode='w+b') as compressed_file_raw:
                    with TextIOWrapper(compressed_file_raw) as compressed_file:
                        json.dump(traceback_info, compressed_file)
            except IOError:
                _logger.error('Unable to compress traceback on disk. Reporting error without traceback', exc_info=True)
                return

            traceback_file.seek(0)
            resp = self.client.api.session.put(traceback_url, data=traceback_file) # pylint: disable=no-member
            resp.raise_for_status()
        error.refresh()

    def add_failure(self, message, **kwargs):
        return self.add_error(message, is_failure=True, **kwargs)

    def _get_id_key(self):
        if type(self).__name__ == 'Test':
            return 'test_id'
        return 'session_id'



from sentinels import NOTHING

from .api_object import APIObject
from .commentable import Commentable
from .error_container import ErrorContainer
from .related_entity_container import RelatedEntityContainer
from .warning_container import WarningContainer
from .lazy_query import LazyQuery
from .metadata_holder import MetadataHolder
from .timing_container import TimingContainer


class Test(APIObject, MetadataHolder, ErrorContainer, WarningContainer, Commentable, RelatedEntityContainer, TimingContainer):

    @property
    def ui_url(self):
        return self.client.get_ui_url('sessions/{}/tests/{}'.format(self.session_display_id, self.logical_id or self.id))

    def report_end(self, duration=NOTHING):
        self.client.api.call_function('report_test_end', {'id': self.id, 'duration': duration})

    def mark_skipped(self, reason=None):
        self.client.api.call_function('report_test_skipped', {'id': self.id, 'reason': reason})

    def report_interrupted(self):
        self.client.api.call_function('report_test_interrupted', {'id': self.id})

    def query_errors(self):
        """Queries tests of the current session

        :rtype: A lazy query object
        """
        return LazyQuery(self.client, '/rest/errors', query_params={'test_id': self.id})

    def get_session(self):
        return self.client.api.get('/rest/sessions/{0}'.format(self.session_id))

    def get_parent(self):
        return self.get_session()

    def update_status_description(self, description):
        return self.client.api.call_function('update_status_description', {'test_id': self.id, 'description': description})



class RelatedEntityContainer(object):

    def add_related_entity(self, entity_type, entity_name):
        # pylint: disable=no-member
        self.client.api.call_function('add_related_entity', {
            self._get_id_key(): self.id,
            'type': entity_type,
            'name': entity_name,
            })

    def _get_id_key(self):
        if type(self).__name__ == 'Test':
            return 'test_id'
        return 'session_id'



class FieldFilter(object):

    def __init__(self, field_name):
        super(FieldFilter, self).__init__()
        self.field_name = field_name
        self._filters = []

    def add_to_url(self, url):
        for operator_name, value in self._filters:
            url = url.add_query_param(self.field_name, '{0}:{1}'.format(operator_name, value))
        return url

def _add_field_proxy_operator_method(operator_name):

    def method(self, other):
        self._filters.append((operator_name, other))  # pylint: disable=protected-access
        return self
    method_name = method.__name__ = '__{0}__'.format(operator_name)
    setattr(FieldFilter, method_name, method)

for _operator_name in ['eq', 'ne', 'lt', 'le', 'gt', 'ge']:
    _add_field_proxy_operator_method(_operator_name)


class _Fields(object):

    def __getattr__(self, name):
        return FieldFilter(name)

FIELDS = _Fields()



from uuid import uuid1
import operator

from flask import Flask, jsonify
from flask_loopback import FlaskLoopback
from sentinels import NOTHING
from urlobject import URLObject as URL

import pytest
from backslash import Backslash
from backslash.lazy_query import LazyQuery
from backslash import FIELDS

# pylint: disable=redefined-outer-name

def test_all_objects(query, num_objects):
    assert [obj.id for obj in query] == list(range(num_objects))


def test_count_objects(query, num_objects):
    assert len(query) == num_objects
    assert len(query._fetched) < num_objects  # pylint: disable=protected-access


def test_getitem(query):
    assert query[20].id == 20


def test_last_item(query, num_objects, page_size):
    assert query[-1].id == num_objects - 1
    assert query._fetched[page_size] is NOTHING  # pylint: disable=protected-access


def test_slicing_not_supported(query):
    with pytest.raises(NotImplementedError):
        query[1:20]  # pylint: disable=pointless-statement


def test_querying_simple_equality(query):
    assert query._url.query == ''  # pylint: disable=protected-access
    query = query.filter(x=1)
    assert query._url.query == 'x=1'  # pylint: disable=protected-access


def test_querying_with_field_queries(query, field_value, operator_name, operator_func):
    query = query.filter(operator_func(FIELDS.field_name, field_value))
    assert query._url.query == 'field_name={0}%3A{1}'.format(  # pylint: disable=protected-access
        operator_name, field_value)

def test_querying_between(query):
    assert query.filter(1 <= FIELDS.x <= 2)._url.query == 'x=ge%3A1&x=le%3A2'  # pylint: disable=protected-access


@pytest.fixture
def query(url, page_size):
    pytest.skip('n/i')
    return LazyQuery(Backslash(url, None), path='/', page_size=page_size)


@pytest.fixture(params=['string_value', 2, 2.5])
def field_value(request):
    return request.param


@pytest.fixture(params=['eq', 'ne', 'gt', 'ge', 'lt', 'le'])
def operator_name(request):
    return request.param


@pytest.fixture
def operator_func(operator_name):
    return getattr(operator, operator_name)


@pytest.fixture
def url(request, flask_app):
    address = str(uuid1())
    returned = URL('http://{0}'.format(address))
    webapp = FlaskLoopback(flask_app)
    webapp.activate_address((address, 80))

    @request.addfinalizer
    def finalize():  # pylint: disable=unused-variable
        webapp.deactivate_address((address, 80))
    return returned


@pytest.fixture
def flask_app():

    app = Flask(__name__)
    app.config['PROPAGATE_EXCEPTIONS'] = True

    @app.route('/')
    def view_objects():  # pylint: disable=unused-variable
        return jsonify({
            'sessions': [{'id': i, 'type': 'session'}
                         for i in range(10)]})
    return app


@pytest.fixture
def page_size():
    return 10


@pytest.fixture
def num_objects():
    return 1000


def test_fake_cursor_count():
    lst = list(range(1000))
    assert FakeCursor(lst).count() == 1000
    assert FakeCursor(lst).offset(10).count() == 990
    assert FakeCursor(lst).offset(30).limit(50).count() == 50
    assert FakeCursor(lst).offset(30).limit(50000).count() == 970


class FakeCursor(object):

    def __init__(self, lst):
        super(FakeCursor, self).__init__()
        self._lst = lst
        self._iterated = False

    def __iter__(self):
        assert not self._iterated
        for item in self._lst:
            yield item

    def limit(self, limit):
        assert not self._iterated, 'Already iterated'
        return FakeCursor(self._lst[:limit])

    def offset(self, offset):
        assert not self._iterated, 'Already iterated'
        return FakeCursor(self._lst[offset:])

    def count(self):
        assert not self._iterated
        return len(self._lst)



import os
import doctest

def test_readme_doctests():
    readme_path = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "README.md"))
    assert os.path.exists(readme_path)
    result = doctest.testfile(readme_path, module_relative=False)
    assert result.failed == 0



from backslash.contrib.utils import add_environment_variable_metadata

def test_environment_metadata():
    environ = {'BACKSLASH_METADATA_x': '2', 'BACKSLASH_METADATA_y': '3', 'unused': '666'}
    metadata = {'x': 'bla'}
    add_environment_variable_metadata(metadata, environ=environ.copy())
    assert metadata == {'x': '2', 'y': '3'}


def test_environment_metadata_nested():
    environ = {'BACKSLASH_METADATA_parent.child': '2', 'BACKSLASH_METADATA_parent.child2': '3', 'BACKSLASH_METADATA_x': '666'}
    metadata = {'x': 'bla'}
    add_environment_variable_metadata(metadata, environ=environ.copy())
    assert metadata == {'x': '666', 'parent': {'child': '2', 'child2': '3'}}



import pytest
from backslash.contrib.utils import normalize_file_path

# pylint: disable=redefined-outer-name

def test_normalize_file_path(file_path, expected_path):
    assert file_path != expected_path
    assert normalize_file_path(file_path) == expected_path


@pytest.fixture
def file_path(file_path_and_expected_path):
    return file_path_and_expected_path[0]

@pytest.fixture
def expected_path(file_path_and_expected_path):
    return file_path_and_expected_path[1]

@pytest.fixture
def file_path_and_expected_path(tmpdir):
    full_path = tmpdir.join('a', 'b', 'c', 'd').ensure(dir=True)
    filename = full_path.join('filename.py')
    tmpdir.join('a', 'b', '.git').ensure(dir=True)
    return filename, 'c/d/filename.py'



import ast
import linecache
import os
import types

from .._compat import PY2

try:
    from slash import config as slash_config
except ImportError:
    slash_config = None


_SPECIAL_DIRS = ('.git', '.hg', '.svn')

_HERE = os.path.abspath('.')

_ALLOWED_ATTRIBUTE_TYPES = [int, str, float]
if PY2:
    _ALLOWED_ATTRIBUTE_TYPES.append(long) # pylint: disable=undefined-variable
_ALLOWED_ATTRIBUTE_TYPES = tuple(_ALLOWED_ATTRIBUTE_TYPES)

_FILTERED_MEMBER_TYPES = [types.MethodType, types.FunctionType, type]
if PY2:
    _FILTERED_MEMBER_TYPES.append(types.UnboundMethodType) # pylint: disable=no-member
    _FILTERED_MEMBER_TYPES.append(types.ClassType) # pylint: disable=no-member
_FILTERED_MEMBER_TYPES = tuple(_FILTERED_MEMBER_TYPES)

_MAX_VARIABLE_VALUE_LENGTH = 100


def normalize_file_path(path):
    path = str(path)
    if path.endswith('.pyc'):
        path = path[:-1]
    if not os.path.isabs(path):
        path = os.path.abspath(os.path.join(_HERE, path))
    dirname = os.path.normpath(path)
    while dirname != '/':
        for special_dir in _SPECIAL_DIRS:
            if os.path.isdir(os.path.join(dirname, special_dir)):
                return os.path.normpath(os.path.relpath(path, dirname))
        dirname = os.path.dirname(dirname)
    return path

def _is_ipython_frame(frame):
    return 'ipython-input' in frame['filename'] and frame['lineno'] == 1

def distill_object_attributes(obj, truncate=True):
    repr_blacklisted_types = _get_repr_blacklisted_types()

    return {attr: value
            if isinstance(value, _ALLOWED_ATTRIBUTE_TYPES) else _safe_repr(value, repr_blacklisted_types, truncate=truncate)
            for attr, value in _iter_distilled_object_attributes(obj)}

def _get_repr_blacklisted_types():
    if slash_config is None:
        return ()

    log_config = slash_config.root.log
    return getattr(log_config, 'repr_blacklisted_types', ())


def distill_slash_traceback(err, line_radius=5):
    repr_blacklisted_types = _get_repr_blacklisted_types()
    returned = _extract_frames(err.traceback, repr_blacklisted_types)
    for frame in returned:
        if _is_ipython_frame(frame):
            commands = frame['locals'].get('In', {}).get('value', None)
            if not commands:
                continue
            try:
                commands = ast.literal_eval(commands)
            except SyntaxError:
                continue
            frame['code_lines_before'], frame['code_line'], frame['code_lines_after'] = _splice_lines(commands, len(commands) -1, line_radius)
            frame['lineno'] = len(commands) - 1
        else:
            lines = linecache.getlines(frame['filename'])
            lineno = frame['lineno']
            frame['code_lines_before'], _, frame['code_lines_after'] = _splice_lines(lines, lineno - 1, line_radius)
    return returned

def _extract_frames(traceback, repr_blacklisted_types):
    returned = traceback.to_list()
    for frame, distilled_dict in zip(traceback.frames, returned):
        if 'locals' not in distilled_dict and hasattr(frame, 'python_frame'): # Slash 1.5.x deprecates the `locals` and `globals` attributes
            distilled_dict['locals'] = _capture_locals(frame.python_frame, repr_blacklisted_types)
            distilled_dict['globals'] = _capture_globals(frame.python_frame, repr_blacklisted_types)
    return returned

def _capture_locals(frame, repr_blacklisted_types):
    if frame is None:
        return None
    return {local_name: {"value": _safe_repr(local_value, repr_blacklisted_types)}
            for key, value in frame.f_locals.items()
            if "@" not in key
            for local_name, local_value in _unwrap_object_variable(key, value)}

def _unwrap_object_variable(var_name, var_value):
    yield var_name, var_value
    if var_name != 'self':
        return

    for attr, value in _iter_distilled_object_attributes(var_value):
        yield 'self.{}'.format(attr), value

def _iter_distilled_object_attributes(obj):
    try:
        obj_dict = getattr(obj, '__dict__', {})
    except Exception:       # pylint: disable=broad-except
        obj_dict = {}

    for attr in obj_dict:
        if attr.startswith('__') and attr.endswith('__'):
            continue
        try:
            value = getattr(obj, attr)
        except Exception:   # pylint: disable=broad-except
            continue
        if isinstance(value, _FILTERED_MEMBER_TYPES):
            continue
        yield attr, value

def _capture_globals(frame, repr_blacklisted_types):
    if frame is None:
        return None
    used_globals = set(frame.f_code.co_names)
    return dict((global_name, {"value": _safe_repr(value, repr_blacklisted_types)})
                for global_name, value in frame.f_globals.items()
                if global_name in used_globals and _is_global_included(value))

def _is_global_included(global_variable):
    if isinstance(global_variable, (types.FunctionType, types.MethodType, types.ModuleType, type)):
        return False
    return True

def _splice_lines(lines, pivot, margin):
    if pivot >= len(lines):
        return ([], "<missing line>", [])
    return (lines[max(0, pivot-margin):pivot],
            lines[pivot],
            lines[pivot+1:pivot+1+margin])

_ENV_VARIABLE_METADATA_PREFIX = 'BACKSLASH_METADATA_'

def add_environment_variable_metadata(metadata, environ=None):
    if environ is None:
        environ = os.environ

    for environment_key, environment_value in environ.items():
        if environment_key.startswith(_ENV_VARIABLE_METADATA_PREFIX):
            _nested_assign(metadata, environment_key[len(_ENV_VARIABLE_METADATA_PREFIX):], environment_value)

def _nested_assign(dictionary, key, value):
    parts = key.split('.')
    for index, part in enumerate(parts):
        if index == len(parts) - 1:
            dictionary[part] = value
        else:
            dictionary = dictionary.setdefault(part, {})

def _safe_repr(value, repr_blacklisted_types, truncate=True):
    if isinstance(value, repr_blacklisted_types):
        returned = "<{!r} object {:x}>".format(type(value).__name__, id(value))

    try:
        returned = repr(value)
    except Exception:  # pylint: disable=broad-except
        return "[Unprintable {0!r} object]".format(type(value).__name__)

    if truncate and len(returned) > _MAX_VARIABLE_VALUE_LENGTH:
        returned = returned[:_MAX_VARIABLE_VALUE_LENGTH - 3] + '...'
    return returned



class MetadataHolder(object):

    def set_metadata(self, key, value):
        self.client.api.call_function('set_metadata', {
            'entity_type': self._data['type'],
            'entity_id': self.id,
            'key': key,
            'value': value
            })

    def set_metadata_dict(self, metadata_dict):
        self.client.api.call_function('set_metadata_dict', {
            'entity_type': self._data['type'],
            'entity_id': self.id,
            'metadata': metadata_dict,
            })


    def get_metadata(self):
        return self.client.api.call_function('get_metadata', {
            'entity_type': self._data['type'],
            'entity_id': self.id,
        })



# pylint: disable=no-member
from sentinels import NOTHING

from .api_object import APIObject
from .archiveable import Archiveable
from .commentable import Commentable
from .error_container import ErrorContainer
from .related_entity_container import RelatedEntityContainer
from .warning_container import WarningContainer
from .lazy_query import LazyQuery
from .metadata_holder import MetadataHolder
from .timing_container import TimingContainer

APPEND_UPCOMING_TESTS_STR = 'append_upcoming_tests'

class Session(APIObject, MetadataHolder, ErrorContainer, WarningContainer, Archiveable, Commentable, RelatedEntityContainer, TimingContainer):

    @property
    def ui_url(self):
        return self.client.get_ui_url('/sessions/{}'.format(self.logical_id or self.id))

    def report_end(self, duration=NOTHING, has_fatal_errors=NOTHING):

        kwargs = {'id': self.id, 'duration': duration, 'has_fatal_errors': has_fatal_errors}
        self.client.api.call_function('report_session_end', kwargs)

    def send_keepalive(self):
        self.client.api.call_function('send_keepalive', {'session_id': self.id})

    def report_test_start(self, name, file_name=NOTHING, class_name=NOTHING, test_logical_id=NOTHING, scm=NOTHING,
                          file_hash=NOTHING,
                          scm_revision=NOTHING, scm_dirty=NOTHING, scm_local_branch=NOTHING, scm_remote_branch=NOTHING,
                          is_interactive=NOTHING, variation=NOTHING, metadata=NOTHING, test_index=NOTHING, parameters=NOTHING):

        params = {
            'session_id': self.id,
            'name': name,
            'scm': scm,
            'file_hash': file_hash,
            'scm_revision': scm_revision,
            'scm_dirty': scm_dirty,
            'scm_local_branch': scm_local_branch,
            'scm_remote_branch': scm_remote_branch,
            'class_name': class_name,
            'file_name': file_name,
            'is_interactive': is_interactive,
            'variation': variation,
            'test_logical_id': test_logical_id,
            'test_index': test_index,
            'parameters': _sanitize_params(parameters),
        }

        if metadata is not NOTHING:
            supports_inline_metadata = (self.client.api.info().endpoints.report_test_start.version >= 2)

            if supports_inline_metadata:
                params['metadata'] = metadata

        returned = self.client.api.call_function(
            'report_test_start', params
        )

        if metadata is not NOTHING and not supports_inline_metadata:
            returned.set_metadata_dict(metadata)

        return returned

    def report_test_distributed(self, test_logical_id):
        self.client.api.call_function('report_test_distributed', {'session_id': self.id, 'test_logical_id': test_logical_id})

    def report_upcoming_tests(self, tests):
        self.client.api.call_function(APPEND_UPCOMING_TESTS_STR,
                                      {'tests':tests, 'session_id':self.id}
                                     )

    def report_in_pdb(self):
        self.client.api.call_function('report_in_pdb', {'session_id': self.id})

    def report_not_in_pdb(self):
        self.client.api.call_function('report_not_in_pdb', {'session_id': self.id})

    def report_interrupted(self):
        if 'report_session_interrupted' in self.client.api.info().endpoints:
            self.client.api.call_function('report_session_interrupted', {'id': self.id})


    def add_subject(self, name, product=NOTHING, version=NOTHING, revision=NOTHING):
        return self.client.api.call_function(
            'add_subject',
            {'session_id': self.id, 'name': name, 'product': product, 'version': version, 'revision': revision})

    def edit_status(self, status):
        return self.client.api.call_function('edit_session_status', {'id': self.id, 'status': status})

    def query_tests(self, include_planned=False):
        """Queries tests of the current session

        :rtype: A lazy query object
        """
        params = None
        if include_planned:
            params = {'show_planned':'true'}
        return LazyQuery(self.client, '/rest/sessions/{0}/tests'.format(self.id), query_params=params)

    def query_errors(self):
        """Queries tests of the current session

        :rtype: A lazy query object
        """
        return LazyQuery(self.client, '/rest/errors', query_params={'session_id': self.id})

    def toggle_investigated(self):
        return self.client.api.call_function('toggle_investigated', {'session_id': self.id})

    def get_parent(self):
        return None


def _sanitize_params(params, max_length=100):
    if params is NOTHING:
        return params

    returned = {}

    for key, value in params.items():
        if value is not None and not isinstance(value, (int, bool, float, str, bytes)):
            value = str(value)
        if isinstance(value, (str, bytes)) and len(value) > max_length:
            remainder = 5
            value = value[:max_length-remainder-3] + '...' + value[-remainder:]

        returned[key] = value
    return returned



from backslash import Backslash


def test_client_get_ui_url():
    assert Backslash('http://bla.com', None).get_ui_url() == 'http://bla.com/#/'
    assert Backslash('http://bla.com', None).get_ui_url('/sessions') == 'http://bla.com/#/sessions'
    assert Backslash('http://bla.com', None).get_ui_url('sessions') == 'http://bla.com/#/sessions'



import pytest
from urlobject import URLObject
import slash
import slash.loader

from backslash.contrib.utils import distill_slash_traceback

# pylint: disable=redefined-outer-name


def test_importing_slash_plugin():
    from backslash.contrib import slash_plugin  # pylint: disable=unused-variable


def test_exception_distilling(traceback):
    assert isinstance(traceback, list)
    assert traceback
    for f in traceback:
        assert f['code_line']


def test_exception_distilling_surrounding_code(traceback):
    frame = traceback[-1]
    assert 'def test_failing():' in [line.strip()
                                     for line in frame['code_lines_before']]
    assert '1 / 0' in frame['code_line']
    assert frame['code_lines_before']
    assert not any('1 / 0' in l for l in frame['code_lines_before'])
    assert frame['code_lines_after']
    assert not any('1 / 0' in l for l in frame['code_lines_after'])


def test_rest_url(installed_plugin, server_url):
    installed_plugin.activate()
    assert installed_plugin.rest_url == server_url.add_path('rest')

def test_webapp_url(installed_plugin, server_url):
    installed_plugin.activate()
    expected = server_url
    if not expected.endswith('/'):
        expected += '/'
    expected += '#/'
    assert installed_plugin.webapp_url == expected


def test_session_webapp_url_no_session(installed_plugin):
    assert installed_plugin.session_webapp_url is None

def test_session_webapp_url_with_session(installed_plugin, server_url):
    installed_plugin.activate()
    with slash.Session() as s:
        url = installed_plugin.session_webapp_url
    assert url == '{}/#/sessions/{}'.format(server_url, s.id)


@pytest.fixture
def traceback(error_result):
    [e] = error_result.get_errors()
    return distill_slash_traceback(e)


@pytest.fixture
def error_result():

    def test_failing():
        1 / 0  # pylint: disable=pointless-statement

    with slash.Session() as s:
        with s.get_started_context():
            slash.runner.run_tests(
                slash.loader.Loader().get_runnables([test_failing]))
            [res] = s.results.iter_test_results()
    return res


@pytest.fixture
def installed_plugin(request, server_url):
    from backslash.contrib import slash_plugin
    plugin = slash_plugin.BackslashPlugin(url=str(server_url), runtoken='blap')

    @request.addfinalizer
    def cleanup():              # pylint: disable=unused-variable
        slash.plugins.manager.uninstall(plugin)

    slash.plugins.manager.install(plugin)
    return plugin


@pytest.fixture
def server_url():
    return URLObject('http://some.backslash.server')



# pylint: disable=no-member

import logbook

from sentinels import NOTHING
from urlobject import URLObject as URL

from .api import API
from .lazy_query import LazyQuery


_logger = logbook.Logger(__name__)


class Backslash(object):

    def __init__(self, url, runtoken):
        super(Backslash, self).__init__()
        if not url.startswith('http'):
            url = 'http://{0}'.format(url)
        self._url = URL(url)
        self.api = API(self, url, runtoken)

    @property
    def url(self):
        return self._url

    def get_ui_url(self, fragment=None):
        returned = str(self.url)
        if not returned.endswith('/'):
            returned += '/'
        if not fragment:
            fragment = '/'
        elif not fragment.startswith('/'):
            fragment = '/' + fragment
        returned += '#{}'.format(fragment)
        return returned


    def toggle_user_role(self, user_id, role):
        return self.api.call_function('toggle_user_role', {'user_id': user_id, 'role': role})

    def get_user_run_tokens(self, user_id):
        return self.api.call_function('get_user_run_tokens', {'user_id': user_id})

    def delete_comment(self, comment_id):
        self.api.call_function('delete_comment', {'comment_id': comment_id})

    def report_session_start(self, logical_id=NOTHING,
                             parent_logical_id=NOTHING,
                             is_parent_session=False,
                             child_id=NOTHING,
                             hostname=NOTHING,
                             total_num_tests=NOTHING,
                             user_email=NOTHING,
                             metadata=NOTHING,
                             keepalive_interval=NOTHING,
                             subjects=NOTHING,
                             infrastructure=NOTHING,
                             ttl_seconds=NOTHING,
    ):
        """Reports a new session starting

        :rtype: A session object representing the reported session
        """
        params = {
            'hostname': hostname,
            'logical_id': logical_id,
            'total_num_tests': total_num_tests,
            'user_email': user_email,
            'metadata': metadata,
            'keepalive_interval': keepalive_interval,
            'subjects': subjects,
            'infrastructure': infrastructure,
            'ttl_seconds': ttl_seconds,
        }
        if parent_logical_id is not None or is_parent_session:
            supports_parallel = (self.api.info().endpoints.report_session_start.version >= 2)

            if supports_parallel:
                params['parent_logical_id'] = parent_logical_id
                params['is_parent_session'] = is_parent_session
                params['child_id'] = child_id
                if child_id is not None:
                    del params['total_num_tests']

        returned = self.api.call_function('report_session_start', params)
        return returned

    def query_sessions(self):
        """Queries sessions stored on the server

        :rtype: A lazy query object
        """
        return LazyQuery(self, '/rest/sessions')

    def query_tests(self):
        """Queries tests stored on the server (directly, not via a session)

        :rtype: A lazy query object
        """
        return LazyQuery(self, '/rest/tests')

    def query(self, path, **kwargs):
        return LazyQuery(self, path, **kwargs)



from .lazy_query import LazyQuery


class Commentable(object):

    def post_comment(self, comment):
        return self.client.api.call_function('post_comment', {
            'comment': comment,
            '{}_id'.format(self.type): self.id
            })

    def get_comments(self):
        return LazyQuery(self.client, '/rest/comments', query_params={'{}_id'.format(self.type): self.id})
